# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.5
dpt_vfs/abstract.py
"""

# pylint: disable=unused-argument

from dpt_runtime import SupportsMixin
from dpt_runtime.exceptions import IOException, NotImplementedException, OperationNotSupportedException, ValueException
from dpt_runtime.io import FileLikeCopyMixin

class Abstract(FileLikeCopyMixin, SupportsMixin):
    """
Provides the abstract VFS implementation for an object.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    TYPE_DIRECTORY = 1 << 1
    """
Directory (or collection like) type
    """
    TYPE_FILE = 1
    """
File type
    """
    TYPE_LINK = 1 << 2
    """
Link type
    """

    # pylint: disable=invalid-name

    __slots__ = ( "__weakref__", ) + FileLikeCopyMixin._mixin_slots_ + SupportsMixin._mixin_slots_
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """

    def __init__(self):
        """
Constructor __init__(Abstract)

:since: v1.0.0
        """

        FileLikeCopyMixin.__init__(self)
        SupportsMixin.__init__(self)
    #

    def __del__(self):
        """
Destructor __del__(Abstract)

:since: v1.0.0
        """

        self.close()
    #

    @property
    def implementing_instance(self):
        """
Returns the implementing instance.

:return: (mixed) Implementing instance or "None"
:since:  v1.0.0
        """

        # pylint: disable=bad-option-value,useless-return

        if (not self.is_valid): raise IOException("VFS object not opened")
        return None
    #

    @property
    def implementing_scheme(self):
        """
Returns the implementing scheme name.

:return: (str) Implementing scheme name
:since:  v1.0.0
        """

        raise NotImplementedException()
    #

    @property
    def is_directory(self):
        """
Returns true if the object is representing a directory (or collection).

:return: (bool) True if directory
:since:  v1.0.0
        """

        return bool(self.type & Abstract.TYPE_DIRECTORY)
    #

    @property
    def is_eof(self):
        """
Checks if the pointer is at EOF.

:return: (bool) True on success
:since:  v1.0.0
        """

        raise OperationNotSupportedException()
    #

    @property
    def is_file(self):
        """
Returns true if the object is representing a file.

:return: (bool) True if file
:since:  v1.0.0
        """

        return bool(self.type & Abstract.TYPE_FILE)
    #

    @property
    def is_link(self):
        """
Returns true if the object is representing a link to another object.

:return: (bool) True if link
:since:  v1.0.0
        """

        return bool(self.type & Abstract.TYPE_LINK)
    #

    @property
    def is_valid(self):
        """
Returns true if the object is available.

:return: (bool) True on success
:since:  v1.0.0
        """

        raise NotImplementedException()
    #

    @property
    def mimetype(self):
        """
Returns the mime type of this VFS object.

:return: (str) VFS object mime type
:since:  v1.0.0
        """

        if (not self.is_valid): raise IOException("VFS object not opened")
        return ("text/directory" if (self.is_directory) else "application/octet-stream")
    #

    @property
    def name(self):
        """
Returns the name of this VFS object.

:return: (str) VFS object name
:since:  v1.0.0
        """

        raise OperationNotSupportedException()
    #

    @property
    def size(self):
        """
Returns the size in bytes.

:return: (int) Size in bytes
:since:  v1.0.0
        """

        raise NotImplementedException()
    #

    @property
    def time_created(self):
        """
Returns the UNIX timestamp this object was created.

:return: (int) UNIX timestamp this object was created
:since:  v1.0.0
        """

        raise OperationNotSupportedException()
    #

    @property
    def time_updated(self):
        """
Returns the UNIX timestamp this object was updated.

:return: (int) UNIX timestamp this object was updated
:since:  v1.0.0
        """

        raise OperationNotSupportedException()
    #

    @property
    def type(self):
        """
Returns the type of this object.

:return: (int) Object type
:since:  v1.0.0
        """

        raise NotImplementedException()
    #

    @property
    def url(self):
        """
Returns the URL of this VFS object.

:return: (str) VFS URL
:since:  v1.0.0
        """

        raise NotImplementedException()
    #

    def close(self):
        """
python.org: Flush and close this stream.

:since: v1.0.0
        """

        raise NotImplementedException()
    #

    def flush(self):
        """
python.org: Flush the write buffers of the stream if applicable.

:since: v1.0.0
        """

        raise OperationNotSupportedException()
    #

    def new(self, _type, vfs_url):
        """
Creates a new VFS object.

:param _type: VFS object type
:param vfs_url: VFS URL

:since: v1.0.0
        """

        raise OperationNotSupportedException()
    #

    def open(self, vfs_url, readonly = False):
        """
Opens a VFS object. The handle is set at the beginning of the object.

:param vfs_url: VFS URL
:param readonly: Open object in readonly mode

:since: v1.0.0
        """

        raise NotImplementedException()
    #

    def read(self, n = 0, timeout = -1):
        """
python.org: Read up to n bytes from the object and return them.

:param n: How many bytes to read from the current position (0 means until
          EOF)
:param timeout: Timeout to use (if supported by implementation)

:return: (bytes) Data; None if EOF
:since:  v1.0.0
        """

        raise OperationNotSupportedException()
    #

    def scan(self):
        """
Scan over objects of a collection like a directory.

:return: (list) Child VFS objects
:since:  v1.0.0
        """

        raise OperationNotSupportedException()
    #

    def seek(self, offset):
        """
python.org: Change the stream position to the given byte offset.

:param offset: Seek to the given offset

:return: (int) Return the new absolute position.
:since:  v1.0.0
        """

        raise OperationNotSupportedException()
    #

    def tell(self):
        """
python.org: Return the current stream position as an opaque number.

:return: (int) Stream position
:since:  v1.0.0
        """

        raise OperationNotSupportedException()
    #

    def truncate(self, new_size):
        """
python.org: Resize the stream to the given size in bytes.

:param new_size: Cut file at the given byte position

:return: (int) New file size
:since:  v1.0.0
        """

        raise OperationNotSupportedException()
    #

    def write(self, b, timeout = -1):
        """
python.org: Write the given bytes or bytearray object, b, to the underlying
raw stream and return the number of bytes written.

:param b: (Over)write file with the given data at the current position
:param timeout: Timeout to use (defaults to construction time value)

:return: (int) Number of bytes written
:since:  v1.0.0
        """

        raise OperationNotSupportedException()
    #

    @staticmethod
    def _get_id_from_vfs_url(vfs_url):
        """
Returns the ID part of the VFS URL given.

:return: (str) VFS URL ID
:since:  v1.0.0
        """

        if (type(vfs_url) is not str): raise ValueException("VFS URL given is invalid")

        vfs_url_data = vfs_url.split("://", 1)
        if (len(vfs_url_data) == 1): raise ValueException("VFS URL '{0}' is invalid".format(vfs_url))

        _return = vfs_url_data[1]

        if (_return in ( "", "/" )): _return = ""
        else: _return = (_return[1:] if (_return[:1] == "/") else _return).strip()

        return _return
    #

    @staticmethod
    def _get_scheme_from_vfs_url(vfs_url):
        """
Returns the scheme of the VFS URL given.

:return: (str) VFS URL scheme
:since:  v1.0.0
        """

        if (type(vfs_url) is not str): raise ValueException("VFS URL given is invalid")

        vfs_url_data = vfs_url.split("://", 1)
        if (len(vfs_url_data) == 1): raise ValueException("VFS URL '{0}' is invalid".format(vfs_url))

        return vfs_url_data[0].lower()
    #
#



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.5
dpt_vfs/implementation.py
"""

from dpt_module_loader import NamedClassLoader
from dpt_runtime import Binary
from dpt_runtime.exceptions import IOException

from .abstract import Abstract

class Implementation(object):
    """
"Implementation" provides implementation independent methods to access VFS
objects.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    TYPE_DIRECTORY = Abstract.TYPE_DIRECTORY
    """
Directory (or collection like) type
    """
    TYPE_FILE = Abstract.TYPE_FILE
    """
File type
    """
    TYPE_LINK = Abstract.TYPE_LINK
    """
Link type
    """

    __slots__ = ( )
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """

    @staticmethod
    def get_class(scheme):
        """
Returns an VFS object class for the given scheme.

:return: (object) VFS object class
:since:  v1.0.0
        """

        _return = NamedClassLoader.get_class_in_namespace("dpt_vfs", "{0}.Object".format(scheme.replace("-", "_")))

        if (_return is None
            or (not issubclass(_return, Abstract))
           ): raise IOException("VFS object not defined for URL scheme '{0}'".format(scheme))

        return _return
    #

    @staticmethod
    def get_instance(scheme):
        """
Returns an VFS object instance for the given scheme.

:return: (object) VFS object instance
:since:  v1.0.0
        """

        vfs_object_class = Implementation.get_class(scheme)
        return vfs_object_class()
    #

    @staticmethod
    def load_vfs_url(vfs_url, readonly = False):
        """
Returns the initialized object instance for the given VFS URL.

:param vfs_url: VFS URL
:param readonly: Open object in readonly mode

:return: (object) VFS object instance
:since:  v1.0.0
        """

        # pylint: disable=protected-access

        vfs_url = Binary.str(vfs_url)
        scheme = Abstract._get_scheme_from_vfs_url(vfs_url)

        _return = Implementation.get_instance(scheme)
        _return.open(vfs_url, readonly)

        return _return
    #

    @staticmethod
    def new_vfs_url(_type, vfs_url):
        """
Returns a new object instance for the given VFS URL.

:param _type: VFS object type
:param vfs_url: VFS URL

:return: (object) VFS object instance
:since:  v1.0.0
        """

        # pylint: disable=protected-access

        vfs_url = Binary.str(vfs_url)
        scheme = Abstract._get_scheme_from_vfs_url(vfs_url)

        _return = Implementation.get_instance(scheme)
        _return.new(_type, vfs_url)

        return _return
    #
#



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.5
dpt_vfs/__init__.py
"""

from .abstract import Abstract
from .abstract_watcher import AbstractWatcher
from .file_like_wrapper_mixin import FileLikeWrapperMixin
from .implementation import Implementation
from .watcher_implementation import WatcherImplementation



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.5
dpt_vfs/dpt_vfs/file/watcher_pyinotify_sync.py
"""

# pylint: disable=import-error

from dpt_logging import LogLine
from pyinotify import Notifier

from .watcher_pyinotify import WatcherPyinotify
from .watcher_pyinotify_callback import WatcherPyinotifyCallback

class WatcherPyinotifySync(WatcherPyinotify):
    """
"file:///" watcher using pyinotify's (synchronous) Notifier.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    # pylint: disable=bad-option-value,slots-on-old-class
    __slots__ = ( )
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """

    def check(self, _path):
        """
Checks a given path for changes if "is_synchronous()" is true.

:param _path: Filesystem path

:return: (bool) True if the given path URL has been changed since last check
         and "is_synchronous()" is true.
:since:  v1.0.0
        """

        with self._lock:
            if (self._pyinotify_instance.check_events()):
                self._pyinotify_instance.read_events()
                self._pyinotify_instance.process_events()
            #
        #

        return False
    #

    def _init_notifier(self):
        """
Initializes the pyinotify instance.

:since: v1.0.0
        """

        with self._lock:
            if (self._pyinotify_instance is None):
                LogLine.debug("{0!r} mode is synchronous", self, context = "dpt_vfs")
                self._pyinotify_instance = Notifier(self, WatcherPyinotifyCallback(self), timeout = 5)
            #
        #
    #

    def stop(self):
        """
Stops all watchers.

:since: v1.0.0
        """

        with self._lock:
            self.free()
            self._pyinotify_instance = None
        #
    #
#



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
unittest
"""

from os import path
from shutil import rmtree
from time import sleep
import os
import unittest

try: from tempfile import TemporaryDirectory
except ImportError:
    from tempfile import mkdtemp

    class TemporaryDirectory(object):
        """
python.org: Create and return a temporary directory.
        """

        def __init__(self, suffix = "", prefix = "tmp", dir = None):
            """
Constructor __init__(TemporaryDirectory)
            """

            self.dir = dir
            self.name = None
            self.prefix = prefix
            self.suffix = suffix
        #

        def __enter__(self):
            """
python.org: Enter the runtime context related to this object.

:return: (str) Temporary directory path
            """

            self.name = mkdtemp(self.suffix, self.prefix, self.dir)
            return self.name
        #

        def __exit__(self, exc, value, tb):
            """
python.org: Exit the runtime context related to this object.

:return: (bool) True to suppress exceptions
            """

            self.cleanup()
            return False
        #

        def cleanup(self, _warn = False):
            """
python.org: The directory can be explicitly cleaned up by calling the cleanup() method.
            """

            rmtree(self.name)
        #
    #
#

try: from urllib.parse import quote_plus
except ImportError: from urllib import quote_plus

from dpt_file import File
from dpt_vfs.dpt_vfs.file.watcher import Watcher

try: from pyinotify import WatchManager
except ImportError: WatchManager = None

class TestVfsFileWatcher(unittest.TestCase):
    """
UnitTest for dpt_vfs.file.Watcher

:since: v1.0.0
    """

    def setUp(self):
        """
python.org: Hook method for setting up the test fixture before exercising
it.
        """

        self.changed_list = [ ]
        self.watcher = Watcher()
    #

    def tearDown(self):
        """
python.org: Hook method for deconstructing the test fixture after testing
it.
        """

        if (self.watcher is not None): self.watcher.disable()
        self.watcher = None
    #

    def changed_callback(self, event_type, url, changed_value = None):
        """
Saves callback URLs for later analysis.

:param event_type: Filesystem watcher event type
:param url: Filesystem URL watched
:param changed_value: Changed filesystem value
        """

        self.changed_list.append({ "event_type": event_type, "url": url, "changed_value": changed_value })
    #

    def process_scenario(self, base_directory, base_url):
        """
Process a simple scenario.

:param event_type: Filesystem watcher event type
:param url: Filesystem URL watched
:param changed_value: Changed filesystem value
        """

        self.assertFalse(self.watcher.is_watched(base_url))
        self.watcher.register(base_url, self.changed_callback)
        self.assertTrue(self.watcher.is_watched(base_url))

        unittest_directory = path.join(base_directory, "unittest")
        unittest_url = base_url + "/unittest"

        os.mkdir(unittest_directory)

        self.assertFalse(self.watcher.is_watched(unittest_url))
        self.watcher.register(unittest_url, self.changed_callback)
        self.assertTrue(self.watcher.is_watched(unittest_url))

        unittest_file = path.join(base_directory, "unittest.txt")
        unittest_url = base_url + "/unittest.txt"

        self.assertFalse(self.watcher.is_watched(unittest_url))
        self.watcher.register(unittest_url, self.changed_callback)
        self.assertTrue(self.watcher.is_watched(unittest_url))

        _file = File()
        _file.open(unittest_file)
        _file.close(False)
    #

    @unittest.skipIf((WatchManager is None), "pyinotify.WatchManager not usable")
    def test_inotify(self):
        """
Tests pyinotify ThreadedNotifier
        """

        self.watcher.set_implementation(Watcher.IMPLEMENTATION_INOTIFY)
        self.assertFalse(self.watcher.is_synchronous)

        with TemporaryDirectory() as base_directory:
            base_url = "file:///{0}".format(quote_plus(base_directory, "/"))
            self.process_scenario(base_directory, base_url)

            # Let the thread process the changes
            sleep(2)
            self.watcher.stop()

            self.assertFalse(self.watcher.is_watched(base_url))
        #

        self.assertEqual(3, len(self.changed_list))

        # mkdir "unittest"
        self.assertEqual(Watcher.EVENT_TYPE_CREATED, self.changed_list[0]['event_type'])
        self.assertEqual(base_url, self.changed_list[0]['url'])
        self.assertEqual("unittest", self.changed_list[0]['changed_value'])

        # create "unittest.txt"
        self.assertEqual(Watcher.EVENT_TYPE_CREATED, self.changed_list[1]['event_type'])
        self.assertEqual(base_url, self.changed_list[1]['url'])
        self.assertEqual("unittest.txt", self.changed_list[1]['changed_value'])

        # modify "unittest.txt"
        self.assertEqual(Watcher.EVENT_TYPE_MODIFIED, self.changed_list[2]['event_type'])
        self.assertEqual("{0}/unittest.txt".format(base_url), self.changed_list[2]['url'])
        self.assertEqual(None, self.changed_list[2]['changed_value'])
    #

    @unittest.skipIf((WatchManager is None), "pyinotify.WatchManager not usable")
    def test_inotify_sync(self):
        """
Tests pyinotify manually triggered Notifier
        """

        self.watcher.set_implementation(Watcher.IMPLEMENTATION_INOTIFY_SYNC)
        self.assertTrue(self.watcher.is_synchronous)

        with TemporaryDirectory() as base_directory:
            base_url = "file:///{0}".format(quote_plus(base_directory, "/"))
            self.process_scenario(base_directory, base_url)
            self.watcher.check(base_url + "/unittest.txt")

            self.watcher.stop()

            self.assertFalse(self.watcher.is_watched(base_url))
        #

        self.assertEqual(3, len(self.changed_list))

        # mkdir "unittest"
        self.assertEqual(Watcher.EVENT_TYPE_CREATED, self.changed_list[0]['event_type'])
        self.assertEqual(base_url, self.changed_list[0]['url'])
        self.assertEqual("unittest", self.changed_list[0]['changed_value'])

        # create "unittest.txt"
        self.assertEqual(Watcher.EVENT_TYPE_CREATED, self.changed_list[1]['event_type'])
        self.assertEqual(base_url, self.changed_list[1]['url'])
        self.assertEqual("unittest.txt", self.changed_list[1]['changed_value'])

        # modify "unittest.txt"
        self.assertEqual(Watcher.EVENT_TYPE_MODIFIED, self.changed_list[2]['event_type'])
        self.assertEqual("{0}/unittest.txt".format(base_url), self.changed_list[2]['url'])
        self.assertEqual(None, self.changed_list[2]['changed_value'])
    #

    def test_mtime(self):
        """
Tests filesystem mtime based, manually triggered check
        """

        self.watcher = Watcher()
        self.watcher.set_implementation(Watcher.IMPLEMENTATION_MTIME)
        self.assertTrue(self.watcher.is_synchronous)

        with TemporaryDirectory() as base_directory:
            base_url = "file:///{0}".format(quote_plus(base_directory, "/"))
            self.process_scenario(base_directory, base_url)

            # The following double call should validate that the synchronized check succeeds only once
            self.watcher.check(base_url + "/unittest.txt")
            self.watcher.check(base_url + "/unittest.txt")

            self.watcher.stop()

            self.assertFalse(self.watcher.is_watched(base_url))
        #

        self.assertEqual(1, len(self.changed_list))

        # modify "unittest.txt"
        self.assertEqual(Watcher.EVENT_TYPE_CREATED, self.changed_list[0]['event_type'])
        self.assertEqual("{0}/unittest.txt".format(base_url), self.changed_list[0]['url'])
        self.assertEqual(None, self.changed_list[0]['changed_value'])
    #
#

if (__name__ == "__main__"):
    unittest.main()
#



# -*- coding: utf-8 -*-

"""
direct Python Toolbox
All-in-one toolbox to encapsulate Python runtime variants
----------------------------------------------------------------------------
(C) direct Netware Group - All rights reserved
https://www.direct-netware.de/redirect?dpt;vfs

This Source Code Form is subject to the terms of the Mozilla Public License,
v. 2.0. If a copy of the MPL was not distributed with this file, You can
obtain one at http://mozilla.org/MPL/2.0/.
----------------------------------------------------------------------------
https://www.direct-netware.de/redirect?licenses;mpl2
----------------------------------------------------------------------------
v1.0.5
dpt_vfs/file_like_wrapper_mixin.py
"""

from dpt_runtime.exceptions import IOException

class FileLikeWrapperMixin(object):
    """
The "FileLikeWrapperMixin" redirects FileIO to the registered wrapped
instance.

:author:     direct Netware Group et al.
:copyright:  direct Netware Group - All rights reserved
:package:    dpt
:subpackage: vfs
:since:      v1.0.0
:license:    https://www.direct-netware.de/redirect?licenses;mpl2
             Mozilla Public License, v. 2.0
    """

    # pylint: disable=assigning-non-slot

    _FILE_WRAPPED_METHODS = ( "flush",
                              "read",
                              "seek",
                              "tell",
                              "truncate",
                              "write"
                            )
    """
File IO methods implemented by an wrapped resource.
    """

    _mixin_slots_ = ( "_wrapped_resource", )
    """
Additional __slots__ used for inherited classes.
    """
    __slots__ = ( )
    """
python.org: __slots__ reserves space for the declared variables and prevents
the automatic creation of __dict__ and __weakref__ for each instance.
    """

    def __init__(self):
        """
Constructor __init__(FileLikeWrapperMixin)

:since: v1.0.0
        """

        self._wrapped_resource = None
        """
Wrapped file-like resource
        """
    #

    def __getattribute__(self, name):
        """
python.org: Called unconditionally to implement attribute accesses for
instances of the class.

:param name: Attribute name

:return: (mixed) Instance attribute
:since:  v1.0.0
        """

        # pylint: disable=protected-access

        if (name == "__class__"
            or name not in self.__class__._FILE_WRAPPED_METHODS
           ): _return = object.__getattribute__(self, name)
        else:
            if (self._wrapped_resource is None): self._open_wrapped_resource()
            if (self._wrapped_resource is None): raise IOException("'{0}' not available for {1!r}".format(name, self))

            _return = getattr(self._wrapped_resource, name)
        #

        return _return
    #

    @property
    def _is_wrapped_resource_open(self):
        """
Returns true if a wrapped resource has been opened for this object.

:return: (bool) True if wrapped resource has been opened
:since:  v1.0.0
        """

        return (self._wrapped_resource is not None)
    #

    def close(self):
        """
python.org: Flush and close this stream.

:since: v1.0.0
        """

        if (self._wrapped_resource is not None):
            try: self._wrapped_resource.close()
            finally: self._wrapped_resource = None
        #
    #

    def _open_wrapped_resource(self):
        """
Opens the wrapped resource once needed.

:since: v1.0.0
        """

        pass
    #

    def _set_wrapped_resource(self, resource):
        """
Sets the wrapped resource for this object.

:param resource: Resource providing the file-like API

:since: v1.0.0
        """

        self._wrapped_resource = resource
    #
#
